/* -*- c -*-

   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ------------------------------------------------------------

   config reader

*/

%{

#define _GNU_SOURCE 1

/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>

#ifdef HAVE_WORDEXP_H
# include <wordexp.h>
#endif


#include "error.h"
#include "xmalloc.h"
#include "confhold.h"
#include "misc.h"
#include "ecurses.h"
#include "keymap.h"
#include "cmd.h"
#include "gettext.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define FIRST_ALLOC 128

#define YY_DECL token_t next_token YY_PROTO ((void))

  typedef enum token {KW_SET      = 300,
                      KW_HOOK     = 301,
                      KW_KEY      = 302,
                      IDENTIFIER  = 303,
                      EOL         = 304,
                      QUOTED      = 305,
                      KEY_SPECIAL = 306,
                      KEY_CONTROL = 307,
                      KEY_SIMPLE  = 308,
                      KEY_VALUE   = 309,
                      TEXT        = 310,
  } token_t;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/

  extern char *stpcpy (char *dest, const char *src);

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

  static int     lineno        = 1;
  static token_t current_token = 0;

  /**
   * this is put in error messages
   */
  static const char *fname = NULL;

  /**
   * these are copies of yytext and yyleng
   */
  static char *token_txt = NULL;
  static int   token_len = 0;

  /**
   * key-value
   */
  static int   token_key = 0;

  /**
   * this is set to 0 if confhold does not have appropriate structures
   * to hold new variable, it may happen if someone defines not registered
   * variable
   */
  static int prepared_to_insert = 1;

  /**
   * this holds field's name that is passed to confhold_insert_value
   */
  static char *field_name = NULL;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

  static int parse (void);

/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

%}

%option noyywrap


ID [a-zA-Z_][a-zA-Z0-9_]+
QT (\'[^\']*\')|(\"[^\"]*\")
SKEY1 (<tab>)|(<esc>)|(<up>)|(<down>)|(<left>)|(<right>)|(<pageup>)
SKEY2 (<pagedown>)|(<backspace>)|(<delete>)|(<insert>)|(<enter>)|(<home>)
SKEY3 (<end>)|(<space>)|(<f[0-9]{1,2}>)
KEY ([^<>\n \t:\\\{\}\'\"])|(\\[^C0-9\n])
VKEY \\[0-9]{3}

%%

^[ \t]*#.* /* ignore */

[ \t]      /* ignore */

^set  return KW_SET;

^key  return KW_KEY;

^hook return KW_HOOK;

^list return KW_SET;

{ID}  return IDENTIFIER;

{QT}  return QUOTED;

{KEY} {
  if (*yytext == '\\')
    token_key = yytext[1];
  else
    token_key = yytext[0];
  return KEY_SIMPLE;
}

{SKEY1}|{SKEY2}|{SKEY3} {
  int f;
  
  if (! strcmp (yytext, "<tab>"))
    token_key = 9;
  else if (! strcmp (yytext, "<esc>"))
    token_key = 27;
  else if (! strcmp (yytext, "<up>"))
    token_key = KEY_UP;
  else if (! strcmp (yytext, "<down>"))
    token_key = KEY_DOWN;
  else if (! strcmp (yytext, "<left>"))
    token_key = KEY_LEFT;
  else if (! strcmp (yytext, "<right>"))
    token_key = KEY_RIGHT;
  else if (! strcmp (yytext, "<pageup>"))
    token_key = KEY_PPAGE;
  else if (! strcmp (yytext, "<pagedown>"))
    token_key = KEY_NPAGE;
  else if (! strcmp (yytext, "<backspace>"))
    token_key = KEY_BACKSPACE;
  else if (! strcmp (yytext, "<delete>"))
    token_key = KEY_DC;
  else if (! strcmp (yytext, "<insert>"))
    token_key = KEY_IC;
  else if (! strcmp (yytext, "<enter>"))
    token_key = '\r';
  else if (! strcmp (yytext, "<home>"))
    token_key = KEY_HOME;
  else if (! strcmp (yytext, "<end>"))
    token_key = KEY_END;
  else if (! strcmp (yytext, "<space>"))
    token_key = ' ';

  f = atoi (yytext + 2);
  token_key = KEY_F (f);

  return KEY_SPECIAL;
}

{VKEY} {
  token_key = strtol (yytext + 1, NULL, 8);
  return KEY_VALUE;
}

\\C[a-z] {
  token_key = yytext[2] - 'a' + 1;
  return KEY_CONTROL;
}

\\\n  lineno++;

\n    lineno++; return EOL;

[^ \t\n:\\<>\{\}]+ return TEXT;

.     return *yytext;

<<EOF>> {
  return EOF;
  yyunput (0, NULL);
}

%%


int
confread_read_file (const char *file)
{
  int ret;
  YY_BUFFER_STATE buffer;
  
  fname = file;
  yyin  = fopen (fname, "r");
  if (yyin == NULL){
    error_ (0, errno, "%s", fname);
    return 1;
  }

  lineno        = 1;
  buffer        = yy_create_buffer (yyin, YY_BUF_SIZE);
  current_token = next_token ();
  ret           = parse ();
  fname         = NULL;
  fclose (yyin);

  if (token_txt)
    xfree (token_txt);
  token_txt = NULL;
  
  return ret;
}


/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static char *
expand (char *str)
{
#ifdef HAVE_WORDEXP_H
  wordexp_t  word_vector;
  int        len = 0;
  int        i;
  char      *result;
  char      *end;

  if (wordexp (str, &word_vector, 0)){
    return NULL;
  }

  for (i = 0; i < word_vector.we_wordc; i++){
    len += strlen (word_vector.we_wordv[i]);
  }

  result = xmalloc (len + 1);
  end    = result;
  for (i = 0; i < word_vector.we_wordc; i++){
    end  = stpcpy (end, word_vector.we_wordv[i]);
  }
  *end   = '\0';
  wordfree (&word_vector);

  return result;
#else
  return xstrdup (str);
#endif
}



static cmd_state_t
name_to_state (char *name)
{
  if (!strcmp (name, "folder"))
    return CMD_LIST;
  if (!strcmp (name, "mail"))
    return CMD_READ_MAIL;
  if (!strcmp (name, "select"))
    return CMD_SELECT_BOX;
  if (!strcmp (name, "fetch"))
    return CMD_FETCH;
  if (!strcmp (name, "send"))
    return CMD_SENDER;
  if (!strcmp (name, "attach"))
    return CMD_ATTACH;
  if (!strcmp (name, "abook"))
    return CMD_ABOOK;
  if (!strcmp (name, "abook_add"))
    return CMD_ABOOK_ADD;
  if (!strcmp (name, "ask"))
    return CMD_ASK;
  if (!strcmp (name, "help"))
    return CMD_HELP;
  return CMD_INVALID;
}


/****************************************************************************
 *    PARSER FUNCTIONS
 ****************************************************************************/

/**
 * Please note, that there is no error-recovery.  First error generates
 * parse error message and stops file analysis.
 */

/*
  GRAMMAR

  <command-list>  ::= <command-list> <command> | <command>
  <command>       ::= <hook-command> | <key-command> | <set-command>
  <hook-command>  ::= "hook" ID ID EOL
  <key-command>   ::= "key" ID <key-specifier> ID EOL
  <set-command>   ::= "set" ID <simple-value> EOL |
                      "set" ID "{" <struct-value> "}" EOL
  <key-specifier> ::= KEY | CONTROL-KEY | SPECIAL | VALUE
  <struct-value>  ::= <struct-value> <field> | <field>
  <field>         ::= ID ":" <simple-value> EOL | EOL
  <simple-value>  ::= ID | QUOTED | TEXT
*/


static char *
token_2_string (int token)
{
  static char token_txt[2] = {'\0', '\0'};
  
  if (token < 255){
    *token_txt = token;
    return token_txt;
  }

  switch (token){

    case KW_SET:
      return "set";

    case KW_KEY:
      return "key";

    case KW_HOOK:
      return "hook";

    case IDENTIFIER:
      return _("identifier");

    case EOL:
      /* it's a nasty hack but this is about end of PREVIOUS line */
      lineno--;
      return _("end of line");

    case QUOTED:
      return _("quoted text");

    case KEY_SPECIAL:
    case KEY_CONTROL:
    case KEY_SIMPLE:
    case KEY_VALUE:
      return _("key specifier");
  }

  return "";
}


static void
parse_error (void)
{
  error_ (0, 0, _("%s:%d: parse error near %s"), fname, lineno,
          token_2_string (current_token));
}



static int
match (token_t token)
{
  int old_token = current_token;

  if (token_txt == NULL){
    token_txt = xmalloc (FIRST_ALLOC + yyleng);
  }
  else if (token_len < yyleng + 2){
    token_len += yyleng + 2;
    xfree (token_txt);
    token_txt = xmalloc (token_len);
  }

  memcpy (token_txt, yytext, yyleng + 1);

  current_token = next_token ();
  return old_token != token;
}




static int
set_field_value (void)
{
  char *txt;
  
  switch (current_token){

    case IDENTIFIER:
    case KEY_SIMPLE:
    case TEXT:
    case QUOTED:
      match (current_token);
      if (! prepared_to_insert)
        return 0;
      txt = expand (token_txt);
      if (txt){
        confhold_insert_value (field_name, txt);
        xfree (txt);
      }
      return 0;

    default:
      return 1;
  }
}



static int
set_field (void)
{
  if (current_token == EOL)
    return match (EOL);
  
  if (match (IDENTIFIER))
    return 1;

  field_name = xstrdup (token_txt);

  if (match (':'))
    return 1;

  if (set_field_value ())
    return 1;

  xfree (field_name);
  field_name = NULL;

  return match (EOL);
}



static int
set_struct_value (void)
{
  do {
    if (set_field ())
      return 1;
  }
  while (current_token == IDENTIFIER || current_token == EOL);

  return 0;
}



static int
set_value (void)
{
  char *txt;
  
  switch (current_token){

    case '{':
      match ('{');
      if (match (EOL))
        return 1;
      if (set_struct_value ())
        return 1;
      return match ('}');

    case IDENTIFIER:
    case QUOTED:
    case TEXT:
    case KEY_SIMPLE:
      match (current_token);
      if (! prepared_to_insert)
        return 0;
      txt = expand (token_txt);
      if (txt){
        confhold_insert_value (NULL, txt);
        xfree (txt);
      }
      return 0;
      
    default:
      return 1;
  }
}



static int
set_command (void)
{
  prepared_to_insert = 1;
  
  if (match (KW_SET))
    return 1;

  if (match (IDENTIFIER))
    return 1;

  if (confhold_prepare_to_insert (token_txt))
    prepared_to_insert = 0;

  if (set_value ())
    return 1;

  return match (EOL);
}



static int
hook_command (void)
{
  exec_t *trigger;
  exec_t *action;
  
  if (match (KW_HOOK))
    return 1;

  if (match (IDENTIFIER))
    return 1;

  trigger = exec_lookup (token_txt);

  if (match (IDENTIFIER))
    return 1;

  action = exec_lookup (token_txt);

  if (trigger && action){
    if (trigger->hook == NULL)
      trigger->hook = hook_create (action->fun);
    else
      hook_add (trigger->hook, action->fun);
  }

  return match (EOL);
}



static int
key_specifier (void)
{
  switch (current_token){

    case KEY_SIMPLE:
    case KEY_VALUE:
    case KEY_SPECIAL:
    case KEY_CONTROL:
      match (current_token);
      return 0;

    default:
      return 1;
  }
}


static int
key_command (void)
{
  cmd_state_t  state;
  exec_t      *exec;
  
  if (match (KW_KEY))
    return 1;

  if (match (IDENTIFIER))
    return 1;

  state = name_to_state (token_txt);

  if (key_specifier ())
    return 1;

  if (match (IDENTIFIER))
    return 1;

  exec = exec_lookup (token_txt);

  if (state != CMD_INVALID && exec){
    keymap_add (keymaps + state, token_key, exec->fun);
  }

  return match (EOL);
}


static int
command (void)
{
  switch (current_token){

    case KW_SET:
      return set_command ();

    case KW_HOOK:
      return hook_command ();

    case KW_KEY:
      return key_command ();

    case EOL:
      match (EOL);
      return 0;

    default:
      return 1;
  }
}


static int
command_list (void)
{
  do {
    if (command ()){
      return 1;
    }
  }
  while (current_token == KW_SET || current_token == KW_HOOK
         || current_token == KW_KEY || current_token == EOL);
  return 0;
}



static int
parse (void)
{
  if (command_list ()){
    parse_error ();
    return 1;
  }
  return 0;
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE confread.l
 *
 ****************************************************************************/
