#!/usr/bin/perl -w

# This function takes 2 arguments.  First is the question (without trailing
# space or ask sign).  Second is the default value.  User will be shown the
# question and asked for a response.  If no answer is supplied, the default
# one will be returned.
sub ask
{
    my $question = shift @_;
    my $default  = shift @_;
    print "$question [$default]: ";
    my $response = <STDIN>;
    chop $response;
    if ($response eq ""){
	return $default;
    }
    return $response;
}

# Takes only 1 argument and replaces leading tilde (~) with users home
# directory.
sub replace_tilde
{
    $_ = shift @_;
    s/^~/$ENV{HOME}/g;
    return $_;
}


sub is_valid_maildir
{
    my $dir = shift @_;
    return -d $dir && -d "$dir/cur" && -d "$dir/new" && -d "$dir/tmp";
}


sub make_maildir
{
    my $dir = shift @_;
    if (-d $dir){
	system ("rm -rf $dir");
    }
    if (-d $dir){
	print "\nCouldn't create $dir...\n\n";
	die;
    }
    mkdir "$dir", 0700;
    mkdir "$dir/new", 0700;
    mkdir "$dir/cur", 0700;
    mkdir "$dir/tmp", 0700;
}



# STEP 1
# Set up root directory of mailbox.
do {
    $root = ask ("Where do you want to store your mailboxes", "~/mail");
    $root = replace_tilde ($root);
    if (-d $root){
	$ok = 1;
    }
    elsif (! -e $root){
	mkdir $root, 0700;
	$ok = 1;
    }
    else {
	print "\nYou have specified an existing file to be a location\n";
	print "of your mailboxes collection.  I'm affraid it's impossible.\n";
	print "Elmo stores your mailboxes in a single directory.\n\n";
	print "Please specify a directory name.\n\n"
    }
} while (! $ok);


# STEP 2
# Start box.
if (is_valid_maildir ("$root/inbox")){
    print "\nThe directory inbox seems to be a valid maildir.\n";
    print "Using inbox as a start directory.\n\n";
    $start = "inbox";
}
elsif (-d "$root/inbox"){
    print "\nThere is an inbox directory under the specified directory,\n";
    print "but it is not a valid maildir.  Maybe another program is using\n";
    print "this directory to store your messages, but using different\n";
    print "format.  You will be asked for a box.  If inbox is your\n";
    print "answer, it will be destroyed.\n\n"
}
elsif (-e "$root/inbox"){
    print "\nThere is a file inbox under the specified directory.\n";
    print "It will be used as a start box but you won't be able to\n";
    print "deliver new mail there.\n\n";
    $start = "inbox";
}

if (! defined ($start)){
    $start = ask ("What is your start box", "inbox");
    make_maildir ($start);
}


# STEP 3
# Trash.
if (is_valid_maildir ("$root/trash")){
    print "\nThe directory trash seems to be a valid maildir.\n";
    print "Using it as a trash.\n\n";
    $trash = "trash";
}
elsif (-d "$root/trash"){
    print "\nThere is a trash directory under the specified directory,\n";
    print "but it is not a valid maildir.  Maybe another program is using\n";
    print "this directory to store your messages, but using different\n";
    print "format.  You will be asked for a box.  If trash is your\n";
    print "answer, it will be destroyed.\n\n";
}
elsif (-e "$root/trash"){
    print "\n There is a file trash under the specified directory.\n";
    print "You can't use it as a trash because elmo won't be able to\n";
    print "move messages there.  You will be asked for a box.  If trash\n";
    print "is your answer, it will be removed.\n\n";
}

if (! defined ($trash)){
    $trash = ask ("Where do you want to move deleted messages", "trash");
    make_maildir ($trash);
}


# STEP 4
# SMTP
$my_name     = ask ("What is your name", "");
$email       = ask ("What is your email", "");
$smtp_server = ask ("What is your smtp server", "");


# STEP 5
# POP3
$pop_server = ask ("What is your pop3 server", "");
$username   = ask ("What is your user name", "");
$password   = ask ("What is your password", "");


mkdir ("$ENV{HOME}/.elmo");
$pop_info_dir = '~/.elmo/pop';

if (-e "$ENV{HOME}/.elmorc"){
    print "\nFound existing .elmorc in your home directory.  The existing\n";
    print "file will be renamed to .elmorc.0.\n\n";
    rename "$ENV{HOME}/.elmorc", "$ENV{HOME}/.elmorc.0";
}

open (CONFIG, ">$ENV{HOME}/.elmorc");

print CONFIG "set mailbox {\n  name: 'my box'\n  root: $root\n";
print CONFIG "  trash: $trash\n}\n\n";

print CONFIG "set smtp_acc {\n  name: $smtp_server\n  server: $smtp_server\n";
print CONFIG "  my_name: '$my_name'\n  email: '$email'\n}\n\n";

print CONFIG "set pop_acc {\n  name: $pop_server\n  server: $pop_server\n";
print CONFIG "  username: $username\n  password: $password\n}\n\n";

print CONFIG "set pop_info_dir $pop_info_dir\n\n";

close (CONFIG);
system ("chmod 0600 $ENV{HOME}/.elmorc");
